@@ -47,7 +47,7 @@ static unsigned int flag_mask; /* mask of currently set flags */
 static unsigned int cur_devs; /* current number of defined devices */
 static int cur_dev; /* device being filled in. If negative, none */
 static int trusted=0; /* is the currently parsed device entry trusted? */
-static unsigned int nr_dev; /* number of devices that the current table can 
+static unsigned int nr_dev; /* number of devices that the current table can
 			       hold */
 struct device *devices; /* the device table */
 static int token_nr; /* number of tokens in line */
@@ -58,8 +58,8 @@ static char default_drive='\0'; /* default drive */
 unsigned int mtools_skip_check=0;
 unsigned int mtools_fat_compatibility=0;
 unsigned int mtools_ignore_short_case=0;
-unsigned int mtools_rate_0=0;
-unsigned int mtools_rate_any=0;
+uint8_t mtools_rate_0=0;
+uint8_t mtools_rate_any=0;
 unsigned int mtools_no_vfat=0;
 unsigned int mtools_numeric_tail=1;
 unsigned int mtools_dotted_dir=0;
@@ -75,7 +75,10 @@ typedef struct switches_l {
     enum {
 	T_INT,
 	T_STRING,
-	T_UINT
+	T_UINT,
+	T_UINT8,
+	T_UINT16,
+	T_UQSTRING
     } type;
 } switches_t;
 
@@ -84,8 +87,8 @@ static switches_t global_switches[] = {
     { "MTOOLS_FAT_COMPATIBILITY", (caddr_t) & mtools_fat_compatibility, T_UINT },
     { "MTOOLS_SKIP_CHECK", (caddr_t) & mtools_skip_check, T_UINT },
     { "MTOOLS_NO_VFAT", (caddr_t) & mtools_no_vfat, T_UINT },
-    { "MTOOLS_RATE_0", (caddr_t) &mtools_rate_0, T_UINT },
-    { "MTOOLS_RATE_ANY", (caddr_t) &mtools_rate_any, T_UINT },
+    { "MTOOLS_RATE_0", (caddr_t) &mtools_rate_0, T_UINT8 },
+    { "MTOOLS_RATE_ANY", (caddr_t) &mtools_rate_any, T_UINT8 },
     { "MTOOLS_NAME_NUMERIC_TAIL", (caddr_t) &mtools_numeric_tail, T_UINT },
     { "MTOOLS_DOTTED_DIR", (caddr_t) &mtools_dotted_dir, T_UINT },
     { "MTOOLS_TWENTY_FOUR_HOUR_CLOCK",
@@ -168,19 +171,52 @@ static switches_t dswitches[]= {
     { "MODE", OFFS(mode), T_UINT },
     { "TRACKS",  OFFS(tracks), T_UINT },
     { "CYLINDERS",  OFFS(tracks), T_UINT },
-    { "HEADS", OFFS(heads), T_UINT },
-    { "SECTORS", OFFS(sectors), T_UINT },
+    { "HEADS", OFFS(heads), T_UINT16 },
+    { "SECTORS", OFFS(sectors), T_UINT16 },
     { "HIDDEN", OFFS(hidden), T_UINT },
     { "PRECMD", OFFS(precmd), T_STRING },
     { "BLOCKSIZE", OFFS(blocksize), T_UINT },
-    { "CODEPAGE", OFFS(codepage), T_UINT }
+    { "CODEPAGE", OFFS(codepage), T_UINT },
+    { "DATA_MAP", OFFS(data_map), T_UQSTRING }
 };
 
-static __inline__ char ch_toupper(char ch)
-{
-    return (char) toupper( (unsigned char) ch);
-}
+#if (defined  HAVE_TOUPPER_L || defined HAVE_STRNCASECMP_L)
+static locale_t C=NULL;
 
+static void init_canon(void) {
+    if(C == NULL)
+	C = newlocale(LC_CTYPE_MASK, "C", NULL);
+}
+#endif
+
+#ifdef HAVE_TOUPPER_L
+static int canon_drv(int drive) {
+    int ret;
+    init_canon();
+    ret = toupper_l(drive, C);
+    return ret;
+}
+#else
+static int canon_drv(int drive) {
+    return toupper(drive);
+}
+#endif
+
+#ifdef HAVE_STRNCASECMP_L
+static int cmp_tok(const char *a, const char *b, size_t len) {
+    init_canon();
+    return strncasecmp_l(a, b, len, C);
+}
+#else
+static int cmp_tok(const char *a, const char *b, size_t len) {
+    return strncasecmp(a, b, len);
+}
+#endif
+
+
+static char ch_canon_drv(char drive) {
+    return (char) canon_drv( (unsigned char) drive);
+}
 
 static void maintain_default_drive(char drive)
 {
@@ -210,10 +246,15 @@ static void syntax(const char *msg, int thisLine)
     fprintf(stderr,"Syntax error at line %d ", lastTokenLinenumber);
     if(drive) fprintf(stderr, "for drive %c: ", drive);
     if(token) fprintf(stderr, "column %ld ", (long)(token - buffer));
-    fprintf(stderr, "in file %s: %s\n", filename, msg);
+    fprintf(stderr, "in file %s: %s", filename, msg);
+    if(errno != 0)
+	fprintf(stderr, " (%s)", strerror(errno));
+    fprintf(stderr, "\n");
     exit(1);
 }
 
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wcast-align"
 static void get_env_conf(void)
 {
     char *s;
@@ -222,20 +263,39 @@ static void get_env_conf(void)
     for(i=0; i< sizeof(global_switches) / sizeof(*global_switches); i++) {
 	s = getenv(global_switches[i].name);
 	if(s) {
-	    if(global_switches[i].type == T_INT)
-		* ((int *)global_switches[i].address) = (int) strtol(s,0,0);
-	    if(global_switches[i].type == T_UINT)
-		* ((unsigned int *)global_switches[i].address) =
-			(unsigned int) strtoul(s,0,0);
-	    else if (global_switches[i].type == T_STRING)
+	    errno = 0;
+	    switch(global_switches[i].type) {
+	    case T_INT:
+		* ((int *)global_switches[i].address) = strtoi(s,0,0);
+		break;
+	    case T_UINT:
+		* ((unsigned int *)global_switches[i].address) = strtoui(s,0,0);
+		break;
+	    case T_UINT8:
+		* ((uint8_t *)global_switches[i].address) = strtou8(s,0,0);
+		break;
+	    case T_UINT16:
+		* ((uint16_t *)global_switches[i].address) = strtou16(s,0,0);
+		break;
+	    case T_STRING:
+	    case T_UQSTRING:
 		* ((char **)global_switches[i].address) = s;
+		break;
+	    }
+	    if(errno != 0) {
+		fprintf(stderr, "Bad number %s for %s (%s)\n", s,
+			global_switches[i].name,
+			strerror(errno));
+		exit(1);
+	    }
 	}
     }
 }
+#pragma GCC diagnostic pop
 
 static int mtools_getline(void)
 {
-    if(!fp || !fgets(buffer, MAX_LINE_LEN, fp))
+    if(!fp || !fgets(buffer, MAX_LINE_LEN+1, fp))
 	return -1;
     linenumber++;
     pos = buffer;
@@ -245,7 +305,7 @@ static int mtools_getline(void)
 	syntax("line too long", 1);
     return 0;
 }
-		
+
 static void skip_junk(int expect)
 {
     lastTokenLinenumber = linenumber;
@@ -281,7 +341,7 @@ static char *get_next_token(void)
 static int match_token(const char *template)
 {
     return (strlen(template) == token_length &&
-	    !strncasecmp(template, token, token_length));
+	    !cmp_tok(template, token, token_length));
 }
 
 static void expect_char(char c)
@@ -307,21 +367,35 @@ static char *get_string(void)
     end = strchr(str, '\"');
     if(!end)
 	syntax("unterminated string constant", 1);
-    *end = '\0';
+    str = strndup(str, ptrdiff(end, str));
     pos = end+1;
     return str;
 }
 
-static unsigned int get_unumber(void)
+static char *get_unquoted_string(void)
+{
+    if(*pos == '"')
+	return get_string();
+    else {
+	char *str=get_next_token();
+	return strndup(str, token_length);
+    }
+}
+
+static unsigned long get_unumber(unsigned long max)
 {
     char *last;
-    unsigned int n;
+    unsigned long n;
 
     skip_junk(1);
     last = pos;
-    n=(unsigned int) strtoul(pos, &pos, 0);
+    n=strtoul(pos, &pos, 0);
+    if(errno)
+	syntax("bad number", 0);
     if(last == pos)
 	syntax("numeral expected", 0);
+    if(n > max)
+	syntax("number too big", 0);
     pos++;
     token_nr++;
     return n;
@@ -335,6 +409,8 @@ static int get_number(void)
     skip_junk(1);
     last = pos;
     n=(int) strtol(pos, &pos, 0);
+    if(errno)
+	syntax("bad number", 0);
     if(last == pos)
 	syntax("numeral expected", 0);
     pos++;
@@ -347,7 +423,7 @@ static void purge(char drive, int fn)
 {
     unsigned int i, j;
 
-    drive = ch_toupper(drive);
+    drive = ch_canon_drv(drive);
     for(j=0, i=0; i < cur_devs; i++) {
 	if(devices[i].drive != drive ||
 	   devices[i].file_nr == fn)
@@ -366,7 +442,7 @@ static void grow(void)
 	}
     }
 }
-	
+
 
 static void init_drive(void)
 {
@@ -392,7 +468,7 @@ static void prepend(void)
 static void append(void)
 {
     grow();
-    cur_dev = cur_devs;
+    cur_dev = (int) cur_devs;
     cur_devs++;
     init_drive();
 }
@@ -419,18 +495,18 @@ static void finish_drive_clause(void)
     }
     devices[cur_dev].file_nr = file_nr;
     devices[cur_dev].cfg_filename = filename;
-    if(! (flag_mask & PRIV_FLAG) && IS_SCSI(&devices[cur_dev]))
-	devices[cur_dev].misc_flags |= PRIV_FLAG;
     if(!trusted && (devices[cur_dev].misc_flags & PRIV_FLAG)) {
 	fprintf(stderr,
 		"Warning: privileged flag ignored for drive %c: defined in file %s\n",
-		toupper(devices[cur_dev].drive), filename);
+		canon_drv(devices[cur_dev].drive), filename);
 	devices[cur_dev].misc_flags &= ~PRIV_FLAG;
     }
     trusted = 0;
     cur_dev = -1;
 }
 
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wcast-align"
 static int set_var(struct switches_l *switches, int nr,
 		   caddr_t base_address)
 {
@@ -438,20 +514,35 @@ static int set_var(struct switches_l *switches, int nr,
     for(i=0; i < nr; i++) {
 	if(match_token(switches[i].name)) {
 	    expect_char('=');
+	    /* All pointers cast back to pointers with alignment
+	     * constraints were such pointers with alignment
+	     * constraints initially, thus they do indeed fit the
+	     * constraint */
+
 	    if(switches[i].type == T_UINT)
 		* ((unsigned int *)((long)switches[i].address+base_address)) =
-		    get_unumber();
-	    if(switches[i].type == T_INT)
+		    (unsigned int) get_unumber(UINT_MAX);
+	    else if(switches[i].type == T_UINT8)
+		* ((uint8_t *)((long)switches[i].address+base_address)) =
+		    (uint8_t) get_unumber(UINT8_MAX);
+	    else if(switches[i].type == T_UINT16)
+		* ((uint16_t *)((long)switches[i].address+base_address)) =
+		    (uint16_t) get_unumber(UINT16_MAX);
+	    else if(switches[i].type == T_INT)
 		* ((int *)((long)switches[i].address+base_address)) =
 		    get_number();
 	    else if (switches[i].type == T_STRING)
 		* ((char**)((long)switches[i].address+base_address))=
-		    strdup(get_string());
+		    get_string();
+	    else if (switches[i].type == T_UQSTRING)
+		* ((char**)((long)switches[i].address+base_address))=
+		    get_unquoted_string();
 	    return 0;
 	}
     }
     return 1;
 }
+#pragma GCC diagnostic pop
 
 static int set_openflags(struct device *dev)
 {
@@ -469,7 +560,6 @@ static int set_openflags(struct device *dev)
 static int set_misc_flags(struct device *dev)
 {
     unsigned int i;
-
     for(i=0; i < sizeof(misc_flags) / sizeof(*misc_flags); i++) {
 	if(match_token(misc_flags[i].name)) {
 	    flag_mask |= misc_flags[i].flag;
@@ -514,7 +604,7 @@ static int set_def_format(struct device *dev)
     return 1;
 }
 
-static int parse_one(int privilege);
+static void parse_all(int privilege);
 
 void set_cmd_line_image(char *img) {
   char *ofsp;
@@ -529,7 +619,7 @@ void set_cmd_line_image(char *img) {
     devices[cur_dev].name = strdup(img);
     devices[cur_dev].offset = 0;
   } else {
-    devices[cur_dev].name = strndup(img, ofsp - img);
+    devices[cur_dev].name = strndup(img, ptrdiff(ofsp, img));
     devices[cur_dev].offset = str_to_offset(ofsp+2);
   }
 
@@ -548,28 +638,56 @@ void set_cmd_line_image(char *img) {
     lastTokenLinenumber = 0;
     pos = buffer;
     token = 0;
-    while (parse_one(0));
+    parse_all(0);
   }
 }
 
+void check_number_parse_errno(char c, const char *oarg, char *endptr) {
+    if(endptr && *endptr) {
+	fprintf(stderr, "Bad number %s\n", oarg);
+	exit(1);
+    }
+    if(errno) {
+	fprintf(stderr, "Bad number %s for -%c (%s)\n", oarg,
+		c, strerror(errno));
+	exit(1);
+    }
+}
+
+static uint16_t tou16(int in, const char *comment) {
+    if(in > (int) UINT16_MAX) {
+	fprintf(stderr, "Number of %s %d too big\n", comment, in);
+	exit(1);
+    }
+    if(in < 0) {
+	fprintf(stderr, "Number of %s %d negative\n", comment, in);
+	exit(1);
+    }
+    return (uint16_t) in;
+}
+
 static void parse_old_device_line(char drive)
 {
     char name[MAXPATHLEN];
     int items;
     long offset;
 
+    int heads, sectors, tracks;
+
     /* finish any old drive */
     finish_drive_clause();
 
     /* purge out data of old configuration files */
     purge(drive, file_nr);
-	
+
     /* reserve slot */
     append();
-    items = sscanf(token,"%c %s %i %ui %ui %ui %li",
+    items = sscanf(token,"%c %s %i %i %i %i %li",
 		   &devices[cur_dev].drive,name,&devices[cur_dev].fat_bits,
-		   &devices[cur_dev].tracks,&devices[cur_dev].heads,
-		   &devices[cur_dev].sectors, &offset);
+		   &tracks,&heads,&sectors, &offset);
+    devices[cur_dev].heads = tou16(heads, "heads");
+    devices[cur_dev].sectors = tou16(sectors, "sectors");
+    devices[cur_dev].tracks = (unsigned int) tracks;
     devices[cur_dev].offset = (off_t) offset;
     switch(items){
 	case 2:
@@ -595,13 +713,14 @@ static void parse_old_device_line(char drive)
 	devices[cur_dev].sectors = 0;
 	devices[cur_dev].heads = 0;
     }
-	
-    devices[cur_dev].drive = ch_toupper(devices[cur_dev].drive);
+
+    devices[cur_dev].drive = ch_canon_drv(devices[cur_dev].drive);
     maintain_default_drive(devices[cur_dev].drive);
     if (!(devices[cur_dev].name = strdup(name))) {
 	printOom();
 	exit(1);
     }
+    devices[cur_dev].misc_flags |= MFORMAT_ONLY_FLAG;
     finish_drive_clause();
     pos=0;
 }
@@ -626,7 +745,7 @@ static int parse_one(int privilege)
 	    syntax("drive letter expected", 0);
 
 	if(action==1 || action==4)
-	    /* replace existing drive */			
+	    /* replace existing drive */
 	    purge(token[0], file_nr);
 	if(action==4)
 	    return 1;
@@ -637,7 +756,7 @@ static int parse_one(int privilege)
 	memset((char*)(devices+cur_dev), 0, sizeof(*devices));
 	trusted = privilege;
 	flag_mask = 0;
-	devices[cur_dev].drive = ch_toupper(token[0]);
+	devices[cur_dev].drive = ch_canon_drv(token[0]);
 	maintain_default_drive(devices[cur_dev].drive);
 	expect_char(':');
 	return 1;
@@ -660,6 +779,12 @@ static int parse_one(int privilege)
     return 1;
 }
 
+static void parse_all(int privilege) {
+    errno=0;
+    while (parse_one(privilege));
+}
+
+
 static int parse(const char *name, int privilege)
 {
     if(fp) {
@@ -683,7 +808,7 @@ static int parse(const char *name, int privilege)
     token = 0;
     cur_dev = -1; /* no current device */
 
-    while(parse_one(privilege));
+    parse_all(privilege);
     finish_drive_clause();
     fclose(fp);
     filename = NULL;
@@ -697,7 +822,7 @@ void read_config(void)
     char *envConfFile;
     static char conf_file[MAXPATHLEN+sizeof(CFG_FILE1)];
 
-	
+
     /* copy compiled-in devices */
     file_nr = 0;
     cur_devs = nr_const_devices;
@@ -746,7 +871,7 @@ void mtoolstest(int argc, char **argv, int type  UNUSEDP)
     char drive='\0';
 
     if(argc > 1 && argv[1][0] && argv[1][1] == ':') {
-	drive = ch_toupper(argv[1][0]);
+	drive = ch_canon_drv(argv[1][0]);
     }
 
     for (dev=devices; dev->name; dev++) {
