@@ -36,7 +36,7 @@
 /**
  * Converts input to shortname
  * @param un unix name (in Unix charset)
- * 
+ *
  * @return 1 if name had to be mangled
  */
 static __inline__ int convert_to_shortname(doscp_t *cp, ClashHandling_t *ch,
@@ -47,12 +47,14 @@ static __inline__ int convert_to_shortname(doscp_t *cp, ClashHandling_t *ch,
 	/* Then do conversion to dn */
 	ch->name_converter(cp, un, 0, &mangled, dn);
 	dn->sentinel = '\0';
+	if (dn->base[0] == '\xE5')
+		dn->base[0] = '\x05';
 	return mangled;
 }
 
 static __inline__ void chomp(char *line)
 {
-	int l = strlen(line);
+	size_t l = strlen(line);
 	while(l > 0 && (line[l-1] == '\n' || line[l-1] == '\r')) {
 		line[--l] = '\0';
 	}
@@ -68,7 +70,7 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 {
 	int mangled;
 
-	/* TODO: Would be nice to suggest "autorenamed" version of name, press 
+	/* TODO: Would be nice to suggest "autorenamed" version of name, press
 	 * <Return> to get it.
 	 */
 #if 0
@@ -78,9 +80,6 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 	if(!opentty(0))
 		return 0;
 
-#define maxsize (isprimary ?  MAX_VNAMELEN+1 : 11+1)
-#define name (isprimary ? argname : shortname)
-
 	mangled = 0;
 	do {
 		char tname[4*MAX_VNAMELEN+1];
@@ -93,12 +92,10 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 		if (isprimary)
 			strcpy(longname, tname);
 		else
-			mangled = convert_to_shortname(cp, 
+			mangled = convert_to_shortname(cp,
 						       ch, tname, shortname);
 	} while (mangled & 1);
 	return 1;
-#undef maxsize
-#undef name
 }
 
 /**
@@ -110,7 +107,7 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 static __inline__ clash_action ask_namematch(doscp_t *cp,
 					     dos_name_t *dosname,
 					     char *longname,
-					     int isprimary, 
+					     int isprimary,
 					     ClashHandling_t *ch,
 					     int no_overwrite,
 					     int reason)
@@ -159,7 +156,8 @@ static __inline__ clash_action ask_namematch(doscp_t *cp,
 	}
 
 	if (!isprimary)
-		name = unix_normalize(cp, name_buffer, dosname);
+		name = unix_normalize(cp, name_buffer,
+				      dosname, sizeof(*dosname));
 	else
 		name = longname;
 
@@ -180,12 +178,12 @@ static __inline__ clash_action ask_namematch(doscp_t *cp,
 		fflush(opentty(1));
 		if (mtools_raw_tty) {
 			int rep;
-			rep = fgetc(opentty(1));			
+			rep = fgetc(opentty(1));
 			fputs("\n", stderr);
 			if(rep == EOF)
 				ans[0] = 'q';
 			else
-				ans[0] = rep;
+				ans[0] = (char) rep;
 		} else {
 			if(fgets(ans, 9, opentty(0)) == NULL)
 				ans[0] = 'q';
@@ -301,9 +299,13 @@ static __inline__ clash_action process_namematch(doscp_t *cp,
 			return NAMEMATCH_SKIP;
 		else
 			return NAMEMATCH_OVERWRITE;
-	default:
+	case NAMEMATCH_NONE:
+	case NAMEMATCH_ERROR:
+	case NAMEMATCH_SUCCESS:
+	case NAMEMATCH_GREW:
 		return NAMEMATCH_NONE;
 	}
+	return action;
 }
 
 static int contains_illegals(const char *string, const char *illegals,
@@ -334,7 +336,7 @@ static int is_reserved(char *ans, int islong)
 		    ((islong && !ans[4]) ||
 		     (!islong && !strncmp(ans+4,"    ",4))))
 			return 1;
-	
+
 	return 0;
 }
 
@@ -369,7 +371,8 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 		reason = RESERVED;
 		ch->use_longname = 1;
 		isprimary = 0;
-	} else if(contains_illegals(dosname->base,short_illegals,11)) {
+	} else if(!ch->is_label &&
+		  contains_illegals(dosname->base,short_illegals,11)) {
 		reason = ILLEGALS;
 		ch->use_longname = 1;
 		isprimary = 0;
@@ -385,18 +388,18 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 					ch->use_longname)) {
 			case -1:
 				return NAMEMATCH_ERROR;
-				
+
 			case 0:
 				return NAMEMATCH_SKIP;
 				/* Single-file error error or skip request */
-				
+
 			case 5:
 				return NAMEMATCH_GREW;
 				/* Grew directory, try again */
-				
+
 			case 6:
 				return NAMEMATCH_SUCCESS; /* Success */
-		}	
+		}
 		match_pos = -2;
 		if (ssp->longmatch > -1) {
 			/* Primary Long Name Match */
@@ -404,7 +407,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 			fprintf(stderr,
 				"Got longmatch=%d for name %s.\n",
 				longmatch, longname);
-#endif			
+#endif
 			match_pos = ssp->longmatch;
 			isprimary = 1;
 		} else if ((ch->use_longname & 1) && (ssp->shortmatch != -1)) {
@@ -440,7 +443,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 	}
 	ret = process_namematch(cp, dosname, longname,
 				isprimary, ch, no_overwrite, reason);
-	
+
 	if (ret == NAMEMATCH_OVERWRITE && match_pos > -1){
 		if((entry.dir.attr & 0x5) &&
 		   (ask_confirmation("file is read only, overwrite anyway (y/n) ? ")))
@@ -448,7 +451,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 		/* Free up the file to be overwritten */
 		if(fatFreeWithDirentry(&entry))
 			return NAMEMATCH_ERROR;
-		
+
 #if 0
 		if(isprimary &&
 		   match_pos - ssp->match_free + 1 >= ssp->size_needed){
@@ -588,7 +591,7 @@ static int _mwrite_one(Stream_t *Dir,
 			case NAMEMATCH_ERROR:
 				return -1;	/* Non-file-specific error,
 						 * quit */
-				
+
 			case NAMEMATCH_SKIP:
 				return -1;	/* Skip file (user request or
 						 * error) */
@@ -614,7 +617,7 @@ static int _mwrite_one(Stream_t *Dir,
 					return -1;
 				}
 				expanded = 1;
-				
+
 				if (dir_grow(Dir, scan.max_entry))
 					return -1;
 				continue;
@@ -623,7 +626,9 @@ static int _mwrite_one(Stream_t *Dir,
 				return write_slots(Dir, &dosname, longname,
 						   &scan, cb, arg,
 						   ch->use_longname);
-			default:
+			case NAMEMATCH_NONE:
+			case NAMEMATCH_AUTORENAME:
+			case NAMEMATCH_QUIT:
 				fprintf(stderr,
 					"Internal error: clash_action=%d\n",
 					ret);
@@ -669,6 +674,7 @@ void init_clash_handling(ClashHandling_t *ch)
 	ch->namematch_default[1] = NAMEMATCH_NONE;
 	ch->name_converter = dos_name; /* changed by mlabel */
 	ch->source = -2;
+	ch->is_label = 0;
 }
 
 int handle_clash_options(ClashHandling_t *ch, char c)
@@ -678,7 +684,7 @@ int handle_clash_options(ClashHandling_t *ch, char c)
 		isprimary = 0;
 	else
 		isprimary = 1;
-	c = tolower(c);
+	c = ch_tolower(c);
 	switch(c) {
 		case 'o':
 			/* Overwrite if primary name matches */
